# Autogenerated by Astropy's setup.py on 2018-03-12 23:50:42
from __future__ import unicode_literals
import datetime

version = "3.0.1"
githash = "d73dccb96099e9da4f6a877661a62c1c884dea62"


major = 3
minor = 0
bugfix = 1

release = True
timestamp = datetime.datetime(2018, 3, 12, 23, 50, 42)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
